/*****************************************************************************
 Base template for vector owner of pointers, it destroys all owned elements
 before efectively remove them from the list

 Classes defined:
   CDimListIO
   CDimVectorOwner
   CDimDequeOwner

 DEFINITION

 Author: Dima V. Fedorov <mailto:dima@dimin.net> <http://www.dimin.net/>
 Copyright (c) 2006 Vision Research Lab, UCSB <http://vision.ece.ucsb.edu>

 History:
   04/19/2006 16:20 - First creation

 Ver : 1
*****************************************************************************/

#ifndef DIM_XVECTORS
#define DIM_XVECTORS
//#pragma message(">>>>>  xvectors: included STL extension")

#include <vector>
#include <deque>
#include <fstream>
#include <algorithm>

//----------------------------------------------------------------------------
// CDimListIO
//----------------------------------------------------------------------------

class CDimListIO
{
public:
  virtual ~CDimListIO() {}

  // I/O - to be implemented on inheritance
  virtual int fromFile(std::ifstream *fs) { fs; return 0; };
  virtual int toFile(std::ofstream *fs) { fs; return 0; };
  virtual int readHeader( std::ifstream *fs ) { fs; return 0; };
  virtual int writeHeader( std::ofstream *fs ) { fs; return 0; };

  // I/O complete
  virtual int fromFile(const std::string *fileName);
  virtual int fromFile(const char* fileName);
  virtual int toFile(const std::string *fileName);
  virtual int toFile(const char* fileName);
};

//----------------------------------------------------------------------------
// CDimVectorOwner
//----------------------------------------------------------------------------

template <class T> 
class CDimVectorOwner: public std::vector<T*>, public CDimListIO
{
public:
  virtual ~CDimVectorOwner() {
    clear();
    std::vector<T*>::~vector();
  }
  virtual CDimVectorOwner* clone() const  { return new CDimVectorOwner(*this); };

  void add( const T& nv ) {
    T *v=0;
    try { v = new T(); } catch (...) { return; };
    if (v == 0) return;
    *v = nv;
    this->push_back( v );
  }

  // make sure to remove all stuff created in memory
  iterator erase( iterator Where ) {
    if (Where != NULL) {
      delete *Where;
      *Where = NULL;    
    }
    return std::vector<T*>::erase( Where );
  }
  
  iterator erase( iterator First, iterator Last ) {
    while(First != Last) {
      delete *First;
      *First = NULL;
      ++First;
    }    
    return std::vector<T*>::erase( First, Last );
  };

  void clear() {
		erase(begin(), end());
    return std::vector<T*>::clear();
	}

	void resize(size_type _Newsize) {	
  	T _Val;
    resize(_Newsize, _Val);
	}

	void resize(size_type _Newsize, T _Val) {	
    if (_Newsize == size()) return;
    if (_Newsize < size()) {
			erase(begin() + _Newsize, end());
      std::vector<T*>::resize(_Newsize);
    }
    else {
      unsigned int n = (unsigned int) _Newsize-size();
      for (unsigned int i=0; i<n; ++i)
        add( _Val );
    }
	}

  //-----------------------------------------
  // I/O
  //-----------------------------------------
  using CDimListIO::fromFile;
  using CDimListIO::toFile;

  int fromFile(std::ifstream *fs) { 
    if (fs == NULL) return -1;
    clear();
    for (unsigned int i=0; i<this->size(); ++i) {
      T v; 
      fs->read((char*)&v, sizeof(T));
      add(v);
      if (fs->fail()) return -1; 
    }
    return (int) this->size() * sizeof(T);
  }

  int toFile(std::ofstream *fs) { 
    if (fs == NULL) return -1;
    for (unsigned int i=0; i<this->size(); ++i) {
      T *v = (*this)[i];
      fs->write((char*)v, sizeof(T));
      if (fs->fail()) return -1; 
    }
    return (int) this->size() * sizeof(T);
  }

protected:
  // restrict access to STL container methods in order to enforce the new interface
  using std::vector<T*>::pop_back;
  using std::vector<T*>::push_back;
  using std::vector<T*>::insert;
};

//----------------------------------------------------------------------------
// CDimDequeOwner
//----------------------------------------------------------------------------

template <class T> 
class CDimDequeOwner: public std::deque<T*>, public CDimListIO 
{
public:
  virtual ~CDimDequeOwner() {
    clear();
    std::deque<T*>::~deque();
  }

  virtual CDimDequeOwner* clone() const  { return new CDimDequeOwner(*this); };

  void add( const T& nv ) {
    T *v=0;
    try { v = new T(); } catch (...) { return; };
    if (v == 0) return;
    *v = nv;
    this->push_back( v );
  }

  // make sure to remove all stuff created in memory
  iterator erase( iterator Where ) {
    if (Where != NULL) {
      delete *Where;
      *Where = NULL;    
    }
    return std::deque<T*>::erase( Where );
  }
  
  iterator erase( iterator First, iterator Last ) {
    while(First != Last) {
      delete *First;
      *First = NULL;
      ++First;
    }    
    return std::deque<T*>::erase( First, Last );
  };

  void clear() {
		erase(begin(), end());
    return std::deque<T*>::clear();
	}

	void resize(size_type _Newsize) {	
  	T _Val;
    resize(_Newsize, _Val);
	}

	void resize(size_type _Newsize, T _Val) {	
    if (_Newsize == size()) return;
    if (_Newsize < size()) {
			erase(begin() + _Newsize, end());
      std::deque<T*>::resize(_Newsize);
    }
    else {
      unsigned int n = _Newsize-size();
      for (unsigned int i=0; i<n; ++i)
        add( _Val );
    }
	}

  //-----------------------------------------
  // I/O
  //-----------------------------------------
  using CDimListIO::fromFile;
  using CDimListIO::toFile;

  int fromFile(std::ifstream *fs) { 
    if (fs == NULL) return -1;
    clear();
    for (unsigned int i=0; i<this->size(); ++i) {
      T v; 
      fs->read((char*)&v, sizeof(T));
      add(v);
      if (fs->fail()) return -1; 
    }
    return (int) this->size() * sizeof(T);
  }

  int toFile(std::ofstream *fs) { 
    if (fs == NULL) return -1;
    for (unsigned int i=0; i<this->size(); ++i) {
      T *v = (*this)[i];
      fs->write((char*)v, sizeof(T));
      if (fs->fail()) return -1; 
    }
    return (int) this->size() * sizeof(T);
  }

protected:
  // restrict access to STL container methods in order to enforce the new interface
  using std::deque<T*>::pop_back;
  using std::deque<T*>::push_back;
  using std::deque<T*>::insert;
};

//----------------------------------------------------------------------------
// CDimVectorMath
//----------------------------------------------------------------------------

template <class T> 
class CDimVectorMath: public std::vector<T>
{
public:
  CDimVectorMath( size_type _Count ): std::vector<T>(_Count) {};
  CDimVectorMath( size_type _Count, const T& _Val ): std::vector<T>(_Count, _Val) {};

  double mean( ) { return mean( begin(), end() ); }
  double mean( iterator First, iterator Last ) {
    double s=0.0;
    unsigned int n=0;
    while (First != Last) {
      s += *First;
      ++First;
      ++n;
    }    
    return s/(double)n;
  }

  double medean( ) { return medean( begin(), end() ); }
  double medean( iterator First, iterator Last ) {
    std::vector<T> tmp;
    tmp.assign( First, Last );
    std::sort( tmp.begin(), tmp.end() );
    if (tmp.size()%2 == 0) {
      return ( tmp[ (unsigned int) (tmp.size()/2.0) ] +
               tmp[ (unsigned int) (tmp.size()/2.0)-1 ] ) / 2.0;
    } else {
      return tmp[ (unsigned int) floor(tmp.size()/2.0) ];
    }
    return 0.0;
  }

};


#endif // DIM_XVECTORS
