/*****************************************************************************
 YAKMEL - Yet Another K-Means Clustering Library
 
 in k-means clustering we are given a set of n data points in d-dimensional 
 space and an integer k, and the problem is to determine a set of k points 
 in d-space, called centers, so as to minimize the mean squared distance 
 from each data point to its nearest center. 

 Defined classes:
   CDimKMeansBase         - main base class, introduces kmeans interface
   CDimThreadedKMeansBase - inherited class introducing threading

 Author: Dima V. Fedorov <mailto:dima@dimin.net> <http://www.dimin.net/>
 Copyright (c) 2006 Vision Research Lab, UCSB <http://vision.ece.ucsb.edu>

 TODO:
   v-fold cross-validation?

 History:
   04/19/2006 16:20 - First creation

 Ver : 1
*****************************************************************************/

#ifndef DIM_KMEANS
#define DIM_KMEANS

//----------------------------------------------------------
// Configuration
//----------------------------------------------------------

//#define KM_SINGLE_THREAD //if defined all multithreaded functionality will be disabled

//----------------------------------------------------------
// Note for initialization and distance:
// the macros are provided for convenience of configuring
// supplied code, the intended way for changes in these
// functions is trough inheritance!
//----------------------------------------------------------

// Initialization, define ONLY ONE!!! default is random
#define KM_INIT_RANDOM
//#define KM_INIT_EVEN

// Default distance function:
//   distance_euclidean
//   distance_euclidean_squared
//   distance_manhattan
//   distance_chebychev
//   distance_emd
#define KM_DISTANCE_FUNCTION distance_euclidean

//----------------------------------------------------------
// Code
//----------------------------------------------------------

#include <cstdlib>
#include <ctime>
#include <cmath>
#include <climits>
#include <cfloat>

// if KM_SINGLE_THREAD is defined all multithreaded functionality will be disabled
#ifndef KM_SINGLE_THREAD
  #include <OpenThreads/Thread>
  #include <OpenThreads/Mutex>
  #include <OpenThreads/Barrier>

  #ifdef WIN32
    #include <windows.h>
    #include <process.h>
    #define getpid() _getpid()
  #else
    #include <unistd.h>
    #include <sched.h>  
  #endif 

  inline void dim_sleep(int milisec) {
    #ifdef WIN32
      Sleep(milisec);
    #else
      sleep(ceil(milisec/1000.0));
    #endif 
  }

#else
  #pragma message(">>>>> K Means: Threading disabled!!!")
#endif // #ifndef KM_SINGLE_THREAD

#include <xtypes.h>
#include <xstring.h>
#include "featureVectors.h"

#include "emd.h"

#ifdef min
#undef min
#endif

#ifdef max
#undef max
#endif

static void kmProgressStdOutProc(long done, long total, const char *descr) {
  if (total == 0) return;
  std::cout << descr << " " << done*100/total << "%\n";
}

inline double pow2( double v ) {
  return v*v;
}

inline int iround(double x) {
  return (int) floor(x + 0.5);
}

//****************************************************************************
// Callbacks
//****************************************************************************

// host shows progress of the executon
// done  - value in a range of [0..total]
// total - range maximum
// descr - any text describing the executed procedure
typedef void (*FDimProgressProc)(long done, long total, const char *descr);

// host shows error of the executon
// val   - error code, not treated at the moment
// descr - any text describing the error
typedef void (*FDimErrorProc)(int val, const char *descr);

// plugin should use this to test is processing should be interrupted...
// If it returns TRUE, the operation should be aborted 
typedef bool (*FDimTestAbortProc)( void );


//****************************************************************************
// CDimKMeansBase
//****************************************************************************

template <class T> 
class CDimKMeansBase {
public:
  CDimKMeansBase( int nclasses, int ndimensions );
  virtual ~CDimKMeansBase( );

  virtual void run( CDimVectorOfFeatures<T> *_data );

  // callbacks - define these to interact with k-means in run time
  FDimProgressProc  progressProc;
  FDimErrorProc     errorProc;
  FDimTestAbortProc testAbortProc; 

  inline CDimVectorOfFeatures<T>   *vec_centroids() { return centroids; }
  inline CDimFeatureVector<int>    *vec_labels()    { return &labels; }
  inline CDimFeatureVector<double> *vec_weights()   { return &weights; }
  
  void setMinError( double new_error ) { min_error = new_error; }
  void setStdOutProgress() { progressProc = kmProgressStdOutProc; }
  void setMaxIterations(unsigned int _max_iterations) { max_iterations = _max_iterations; }

protected:
  //---------------------------------------------------------------------------------------
  // functions that do the work and can be changed in inherited classes (polymorphism)
  //---------------------------------------------------------------------------------------
  
  // the main function that does the clustering
  virtual void do_clustering();
  // distance function
  virtual inline double get_distance( CDimFeatureVector<T> *item1, CDimFeatureVector<T> *item2 );
  // initialization function
  virtual int  initialize_centroids();
  virtual void compute_perpoint_distance_distr();
  virtual void sort_classes_by_max_prob( );

protected:
  //---------------------------------------------------------------------------------------
  // internals of the k-means class
  //---------------------------------------------------------------------------------------

  bool is_running;
  unsigned int k;
  unsigned int num_dims;
  unsigned int max_iterations; // the maximum limit for a number of iterations
  double       min_error;      // one of stop conditions, default minimal error is 0.00001

  CDimVectorOfFeatures<T>     *centroids;
  CDimVectorOfFeatures<T>     *data;
  CDimFeatureVector<int>       labels;
  CDimFeatureVector<double>    weights;
  CDimVectorOfFeatures<uchar> *distributions;

  // CALLBACK INVOCATION
  inline void doProgressProc(long done, long total, const char *descr) {
    if (progressProc != NULL) progressProc( done, total, descr);
  }
  inline void doErrorProc(int val, const char *descr) {
    if (errorProc != NULL) errorProc( val, descr);
  }
  inline bool doTestAbortProc( void ) {
    if (testAbortProc != NULL) return testAbortProc( );
  }

  // MISC.
  inline void assignVector(CDimFeatureVector<T> *dst, CDimFeatureVector<T> *src ) {
    unsigned int size = (unsigned int) src->size();
    register unsigned int i;
    for (i=0; i<size; ++i)
      (*dst)[i] = (*src)[i];
  }
};


template<class T>
CDimKMeansBase<T>::CDimKMeansBase( int nclasses, int ndimensions )
    : is_running(false)
{
  k              = (unsigned int) nclasses;
  num_dims       = (unsigned int) ndimensions;
  max_iterations = 0;
  min_error      = 0.00001;

  progressProc = NULL;
  errorProc = NULL;
  testAbortProc = NULL; 

  // now init data vectors
  data      = new CDimVectorOfFeatures<T> (num_dims);
  centroids = new CDimVectorOfFeatures<T> (num_dims, k);
  distributions = NULL;

  weights.resize(num_dims, 1.0);
  weights.fill(1.0);
}

template<class T>
CDimKMeansBase<T>::~CDimKMeansBase( ) {
  centroids->clear();
  data->clear();
  delete centroids;
  delete data;
  if (distributions != NULL) { distributions->clear(); delete distributions; }
  labels.clear();
  weights.clear();
}

template<class T>
void CDimKMeansBase<T>::run( CDimVectorOfFeatures<T> *_data ) {
  if ( (_data == NULL) || (_data->empty()) ) return;

  // convert data
  // we'd rather have fast index access to data points, so copy into internal vector
  data->reserve( _data->size() );

  //data->assign( _data->begin(), _data->end() );  
  unsigned int i;
  for (i=0; i<_data->size(); ++i) {
    data->add( (*_data)[i] );
  }

  // start clustering
  do_clustering();
}

template<class T>
void CDimKMeansBase<T>::do_clustering(  ) {

  unsigned int i, h, j, iter=0;
  unsigned int n = (unsigned int) data->size();
  unsigned int m = num_dims;
  double old_error=0.0, error=DBL_MAX; // sum of squared euclidean distance
  double t = min_error; //minimal error
  unsigned int changed_max = 0;
  xstring progress_str;

  // output cluster label for each data point
  labels.resize(n, 0);
  labels.fill(0);

  // size of each cluster
  CDimFeatureVector<int> counts(k, 0);
  CDimVectorOfFeatures<T> centroidsTemp(num_dims, k);

  is_running = true;
  progress_str.sprintf("Initializing %d centroids", k);
  doProgressProc(0, 100, progress_str.c_str());
  initialize_centroids();

  doProgressProc(0, 100, "Clustering");
  // start iterations
  while (fabs(error - old_error) > t) {
  //while (1) {
    
    // save error from last step
    old_error = error, error = 0;
    unsigned int changed = 0;

    // clear old counts and temp centroids
    counts.fill(0);
    centroidsTemp.fill(0.0f);

    for (h=0; h<n; ++h) {

      // identify the closest cluster
      double min_distance = DBL_MAX;
      int plab = labels[h];
      for (i=0; i<k; ++i) {
        double distance = get_distance( (*data)[h], (*centroids)[i] );
        if (distance < min_distance) {
          plab = i;
          min_distance = distance;
        }
      }
      if (labels[h] != plab) { ++changed; labels[h] = plab; }
    
      // update size and temp centroid of the destination cluster
      for (j=0; j<m; ++j)
        (*centroidsTemp[labels[h]])[j] += (*(*data)[h])[j];
      ++counts[labels[h]];

      // update standard error 
      error += min_distance;
    } // for h

    for (i=0; i<k; ++i) { // update all centroids
      for (j=0; j<m; ++j) {
        (*(*centroids)[i])[j] = counts[i] ? (*centroidsTemp[i])[j] / counts[i] : (*centroidsTemp[i])[j];
      }
    }
    
    progress_str.sprintf("%d #%d", iter, changed);
    if (iter>0) {
      if (changed_max<changed) changed_max = changed;
      doProgressProc( changed_max-changed, changed_max, progress_str.c_str() );
    } else
      doProgressProc( 0, 100, progress_str.c_str() );

    if (changed == 0) break;
    if ( (max_iterations>0) && (iter>max_iterations) ) break;
    ++iter;
  } // while (fabs(error - old_error) > t)

  is_running = false;
}

template<class T>
void CDimKMeansBase<T>::compute_perpoint_distance_distr(  ) {
  unsigned int n = (unsigned int) data->size();
  if (distributions == NULL) 
    distributions = new CDimVectorOfFeatures<unsigned char> (k, n);

  std::vector<double> distances(k);
  unsigned int i, j;
  // iterate through all data points
  for (i=0; i<n; ++i) {

    // for each centroid get distance to it from current data point
    double total_dist=0.0, maxv=0.0, minv=DBL_MAX;
    for (j=0; j<k; ++j) {
      distances[j] = get_distance( (*data)[i], (*centroids)[j] );
      if (maxv < distances[j]) maxv = distances[j];
      if (minv > distances[j]) minv = distances[j];
    }

    // get total distance and relative dist to each centroid
    for (j=0; j<k; ++j) {
      distances[j] = maxv - (distances[j] - minv);
      total_dist += distances[j];
    }

    for (j=0; j<k; ++j) {
      (*(*distributions)[i])[j] = (unsigned char) iround( (distances[j]/total_dist) * 100.0 );
    }

  } // for all data points
}

template<class T>
void CDimKMeansBase<T>::sort_classes_by_max_prob(  ) {
  
  unsigned int n = (unsigned int) data->size();
  unsigned int i, j;
  compute_perpoint_distance_distr( );
  
  // now obtain the average max probability for each class and sort accordingly 
  std::vector<double> probs(k, 0.0);

  for (j=0; j<k; ++j) {
    for (i=0; i<n; ++i) {
      unsigned int total = 0;
      if (labels[i]==j) {
        probs[j] += (*(*distributions)[i])[j];
        ++total;
      }
      probs[j] = probs[j] / total;
    } // i
  } // j

  // now rearrange centroids and labels according to a probs order
  // ... to be continued IF NEEDED
}

#ifdef KM_INIT_EVEN

// initialization - even
template<class T>
int CDimKMeansBase<T>::initialize_centroids() {
  unsigned int i, h;
  unsigned int n = (unsigned int) data->size();
  unsigned int m = num_dims;
  unsigned int hstep = (unsigned int) floor( (n-1.0)/k );

  for (h=i=0; i<k; h+=hstep, ++i)
    assignVector((*centroids)[i], (*data)[h] );

  #ifdef _DEBUG 
  //centroids->toFile("matlab\\centroids_initial.dat");
  #endif
  return 0;
}

#else

// Default initialization - random with minimum distance between centroids
// minimum distance is computed by: max-min/k 
// max and min are computed by measuring the distance from the random element to others
template<class T>
int CDimKMeansBase<T>::initialize_centroids() {
  unsigned int i, h, q, a=0;
  unsigned int n = (unsigned int) data->size();
  unsigned int m = num_dims;
  double min_distance = 0.0, dist;
  unsigned int d;

  // randomize the rand generator, using time as a seed
  srand( (unsigned)time( NULL ) );

  CDimVectorMath<double> min_dists(10);
  CDimFeatureVector<T> v0(m);

  // define minimum distance between centroids
  for (d=0; d<10; ++d) {
    double maxv=0.0, minv=DBL_MAX, newv;
    
    h = (unsigned int) ( (((double)rand())/RAND_MAX)*(n-1.0) );
    assignVector(&v0, (*data)[h] );
    //assignVector(&v0, (*data)[0] );

    for(i=1; i<n; ++i) {
      if (i == h) continue;
      newv = get_distance( (*data)[i], &v0 );

      if (newv > maxv) maxv = newv; 
      else 
      if (newv < minv) minv = newv;
    }
    min_dists[d] = (maxv - minv) / (k);
  } // for d
  //min_distance = min_dists.medean();
  min_distance = min_dists.mean();

  #ifdef _DEBUG 
  std::cout << "\nMinimum dist: " << min_distance << "\n";
  #endif

  // now get random centroids
  i=0;
  while (i<k) {
	  // get random position
    h = (unsigned int) ( (((double)rand())/RAND_MAX)*(n-1.0) );

    // copy texture 
    assignVector((*centroids)[i], (*data)[h] );

    // test if all the distances are above the minimum threshold
	  for (q=0; q<i; ++q) {
      dist = get_distance( (*centroids)[i], (*centroids)[q] );

      if ( dist < min_distance ) {
		    --i;

        #ifdef _DEBUG 
        std::cout << dist << " ";
        #endif

		    break;
      }
	  } // for q<=i

	  ++i;
    
    ++a;
    if (a > n*2) break;
  } // while i<k
  if (a>n*2) initialize_centroids();

  #ifdef _DEBUG 
  std::cout << "\nInitialized " << k << " centroids\n";
  //centroids->toFile("matlab\\centroids_initial.dat");
  #endif
  return 0;
}

#endif // KM_INIT_

//****************************************************************************
// Distances, the best way is to change the distance measure by inheriting
// the class and tailoring it to your particular needs
// there are some default distance measures provided here
//****************************************************************************

//----------------------------------------------------------------------------
// Euclidean - default
// distance(x,y) = { SOMi[Wi(xi - yi)^2] }^1/2
// note: geometric distance in the multidimensional space
//----------------------------------------------------------------------------

template<class T> 
inline double distance_euclidean(CDimFeatureVector<T> *item1, CDimFeatureVector<T> *item2, 
                                 CDimFeatureVector<double> *weights, uint num_dims) {
  double distance = 0.0;
  for (unsigned int i=0; i<num_dims; ++i)
    distance += (*weights)[i] * pow2( ((double)(*item1)[i] - (double)(*item2)[i]) );
        
	return sqrt(distance);
}

//----------------------------------------------------------------------------
// Squared Euclidean
// distance(x,y) = SOM_i[Wi(xi - yi)^2]
// note: place progressively greater weight on objects that are further apart
//----------------------------------------------------------------------------
template<class T> 
inline double distance_euclidean_squared(CDimFeatureVector<T> *item1, CDimFeatureVector<T> *item2, 
                                         CDimFeatureVector<double> *weights, uint num_dims) {
  double distance = 0.0;
  for (unsigned int i=0; i<num_dims; ++i)
    distance += (*weights)[i] * pow2( ((double)(*item1)[i] - (double)(*item2)[i]) );
        
	return distance;
}

//----------------------------------------------------------------------------
// City-block (aka Manhattan) == L1
// distance(x,y) = SOM_i[Wi|xi - yi|]
// note: effect of single large outliers is dampened comparing to eucledian
//----------------------------------------------------------------------------
template<class T> 
inline double distance_manhattan(CDimFeatureVector<T> *item1, CDimFeatureVector<T> *item2, 
                                 CDimFeatureVector<double> *weights, uint num_dims) {
  double distance = 0.0;
  for (unsigned int i=0; i<num_dims; ++i)
    distance += (*weights)[i] * fabs( ((double)(*item1)[i] - (double)(*item2)[i]) );
        
	return (distance);
}

//----------------------------------------------------------------------------
// Chebychev 
// distance(x,y) = max( Wi|xi - yi| )
// note: define objects as "different" if they are different on any one of the dimensions
//----------------------------------------------------------------------------
template<class T> 
inline double distance_chebychev(CDimFeatureVector<T> *item1, CDimFeatureVector<T> *item2, 
                                 CDimFeatureVector<double> *weights, uint num_dims) {
  double distance=0.0;
  for (unsigned int i=0; i<num_dims; ++i) {
    distance = dim::max( distance, (*weights)[i] * fabs((double)(*item1)[i] - (double)(*item2)[i]) );
  }

	return (distance);
}

//----------------------------------------------------------------------------
// EMD - Earth Movers Distance
// 
// note:
//----------------------------------------------------------------------------

template<class T> 
float emd_dist(feature_t *f1, int i1, feature_t *f2, int i2) {
  CDimFeatureVector<T> *item1 = (CDimFeatureVector<T> *) f1;
  CDimFeatureVector<T> *item2 = (CDimFeatureVector<T> *) f2;

  double distance = pow2( ((double)(*item1)[i1] - (double)(*item2)[i2]) );
	return (float) sqrt(distance);
}

template<class T> 
float emd_weight(weight_t *Weights, int pos) {
  CDimFeatureVector<double> *w = (CDimFeatureVector<double> *) Weights;
  return (float) (*w)[pos];
}

template<class T> 
inline double distance_emd(CDimFeatureVector<T> *item1, CDimFeatureVector<T> *item2, 
                           CDimFeatureVector<double> *weights, uint num_dims) {
  signature_t s1 = { num_dims, (void*) item1, (void*) weights };
  signature_t s2 = { num_dims, (void*) item2, (void*) weights };

	return (double) emd(&s1, &s2, emd_dist<T>, emd_weight, 0, 0);
}


// default is euclidean distance
template<class T>
inline double CDimKMeansBase<T>::get_distance( CDimFeatureVector<T> *item1, CDimFeatureVector<T> *item2 ) {
  return KM_DISTANCE_FUNCTION(item1, item2, &weights, num_dims);
}

//****************************************************************************
// KMeansThread
// the thread receive all vectors and a range, it computes min distances
// in this range and creates p/ thread error and changed
// It is by construction of K-Means algorithm each thread only accesses it's
// part of global arrays and thus it's legal!
//****************************************************************************

#ifndef KM_SINGLE_THREAD

template <class T> class CDimThreadedKMeansBase;

template <class T> 
class KMeansThread: public OpenThreads::Thread {
public:
  // b - begin and e - end = range where to compute min distance, both values are included
  KMeansThread (CDimThreadedKMeansBase<T> *_kmb, unsigned int _b, unsigned int _e);
  ~KMeansThread();

  virtual void run();

  inline double       error()   { return _error; }
  inline unsigned int changed() { return _changed; }
  //inline void         setRange(unsigned int _b, unsigned int _e) { rBeg = _b; rEnd = _e; };

  inline CDimVectorOfFeatures<T> *v_centroids_temp()   { return centroidsT; }
  inline CDimFeatureVector<int>  *v_counts() { return counts; }

private:
  CDimThreadedKMeansBase<T> *kmb;
  
  CDimVectorOfFeatures<T> *data;
  CDimVectorOfFeatures<T> *centroids;
  CDimFeatureVector<int>  *labels;
  
  // private arrays
  CDimVectorOfFeatures<T> *centroidsT;
  CDimFeatureVector<int>  *counts;

  unsigned int rBeg, rEnd;
  double _error;
  unsigned int _changed;
  virtual void find_closest_cenroids_in_range( unsigned int b, unsigned int e );
};

template<class T>
KMeansThread<T>::KMeansThread( CDimThreadedKMeansBase<T> *_kmb, unsigned int _b, unsigned int _e )
  : OpenThreads::Thread(), kmb(_kmb), rBeg(_b), rEnd(_e)
{
  data       =  _kmb->data;
  centroids  =  _kmb->centroids;
  labels     = &_kmb->labels;

  // now init data vectors
  counts     = new CDimFeatureVector<int> (_kmb->k, 0);
  centroidsT = new CDimVectorOfFeatures<T> (_kmb->num_dims, _kmb->k);
}

template<class T>
KMeansThread<T>::~KMeansThread() {
  counts->clear();
  centroidsT->clear();
  delete counts;
  delete centroidsT;
}

template<class T>
void KMeansThread<T>::run()
{
  kmb->bar.block(kmb->global_num_threads);  // Sync the threads
  find_closest_cenroids_in_range( rBeg, rEnd );
  kmb->bar.block(kmb->global_num_threads);  // Sync the threads
}

template<class T>
void KMeansThread<T>::find_closest_cenroids_in_range( unsigned int b, unsigned int e ) {
  
  _error = 0;
  _changed = 0;
  unsigned int i, h, j;
  unsigned int k = kmb->k;
  unsigned int m = kmb->num_dims;

  // clear old counts and temp centroids
  counts->fill(0);
  centroidsT->fill(0.0f);

  for (h=b; h<=e; ++h) {

    // identify the closest cluster
    double min_distance = DBL_MAX;
    int plab = (*labels)[h];
    for (i=0; i<k; ++i) {
      double distance = kmb->get_distance( (*data)[h], (*centroids)[i] );
      if (distance < min_distance) {
        plab = i;
        min_distance = distance;
      }
    }
    if ((*labels)[h] != plab) { ++_changed; (*labels)[h] = plab; }
  
    // update size and temp centroid of the destination cluster
    for (j=0; j<m; ++j)
      (*(*centroidsT)[(*labels)[h]])[j] += (*(*data)[h])[j];
    ++(*counts)[(*labels)[h]];

    // update standard error 
    _error += min_distance;
  } // for h
}
#endif

//****************************************************************************
// CDimThreadedKMeansBase
//****************************************************************************

#ifdef KM_SINGLE_THREAD 
// if no threading support is defined then create a class stub

template <class T> 
class CDimThreadedKMeansBase: public CDimKMeansBase<T> {
public:
  CDimThreadedKMeansBase( int nthreads, int nclasses, int ndimensions )
    : CDimKMeansBase<T>(nclasses, ndimensions) {};
};

#else // now the actual threaded code

template <class T> 
class CDimThreadedKMeansBase: public CDimKMeansBase<T> {
public:
  CDimThreadedKMeansBase( int nthreads, int nclasses, int ndimensions );
  ~CDimThreadedKMeansBase( );

protected:
  unsigned int num_threads;
  unsigned int global_num_threads; 
  OpenThreads::Barrier bar;

  void do_clustering();

  friend class KMeansThread<T>;
};


template<class T>
CDimThreadedKMeansBase<T>::CDimThreadedKMeansBase( int nthreads, int nclasses, int ndimensions )
: CDimKMeansBase<T>(nclasses, ndimensions) {

  num_threads = (unsigned int) nthreads;
  global_num_threads = 0;
}

template<class T>
CDimThreadedKMeansBase<T>::~CDimThreadedKMeansBase( ) {

}

template<class T>
void CDimThreadedKMeansBase<T>::do_clustering(  ) {
  
  if (num_threads < 2) {
    CDimKMeansBase<T>::do_clustering( );
    return;
  }

  unsigned int i, j, iter=0;
  unsigned int n = (unsigned int) data->size();
  unsigned int m = num_dims;
  double old_error=0.0, error=DBL_MAX; // sum of squared euclidean distance
  double t = min_error; //minimal error
  unsigned int changed_max = 0;
  xstring progress_str;

  // output cluster label for each data point
  labels.resize(n, 0);

  is_running = true;
  doProgressProc(0, 100, "Initializing centroids");
  initialize_centroids();

  doProgressProc(0, 100, "Clustering");

  CDimFeatureVector<int> counts(k, 0);
  CDimVectorOfFeatures<T> centroidsTemp(num_dims, k);

  //--------------------------------------------
  // create threads
  //--------------------------------------------
  if (num_threads >= n) return;
  std::vector<KMeansThread<T> *> threads;
  threads.reserve(num_threads);

  global_num_threads = num_threads + 1; 
  OpenThreads::Thread::SetConcurrency(num_threads);
  OpenThreads::Thread::Init();

  // create threads
  unsigned int a = (unsigned int) ceil( ((double)n)/((double)num_threads) );
  unsigned int b = 0, e=a-1;
  for (i=0; i<num_threads; ++i) {
    KMeansThread<T> *thread = new KMeansThread<T> (this, b, e);
    //thread->addObserver(&observer);
    //thread->setStackSize(1024*256);
    threads.push_back(thread);
    b += a;
    e += a; if (e>=n) e=n-1;
  }

  //--------------------------------------------
  // start iterations
  //--------------------------------------------
  while (fabs(error - old_error) > t) {
  //while (1) {
    
    // save error from last step
    old_error = error, error = 0;
    unsigned int changed = 0;

    // clear old counts and temp centroids
    counts.fill(0);
    centroidsTemp.fill(0.0f);

    // start threads
    for (i=0; i<num_threads; ++i) {
      KMeansThread<T> *thread = threads[i];
      thread->start();
    }   

    // now wait threads to finish
    bar.block(global_num_threads);  // Block 'till ready
    bar.block(global_num_threads);  // Block 'till finished 

    // all dead?
    while (1) {
      int tr = 0;
      for (i=0; i<num_threads; ++i)
        if (threads.at(i)->isRunning()) {
          ++tr;
          #ifdef _DEBUG
          std::cout << ">>> waiting for: " << i << "\n";
          #endif
        }
      if (tr == 0) break;    
      dim_sleep(10);
    }

    // update state using thread outputs
    for (i=0; i<num_threads; ++i) {
      KMeansThread<T> *thread = threads[i];
      error += thread->error();
      changed += thread->changed();
      
      // summ up all the counts and temp centroids
      CDimFeatureVector<int>  *cThread   = thread->v_counts();
      CDimVectorOfFeatures<T> *cctThread = thread->v_centroids_temp();
      
      for (i=0; i<k; ++i) {
        counts[i] += (*cThread)[i];
        for (j=0; j<m; ++j) {
          (*centroidsTemp[i])[j] += (*(*cctThread)[i])[j];
        } // for j
      } // for i
    }   

    for (i=0; i<k; ++i) { // update all centroids
      for (j=0; j<m; ++j) {
        (*(*centroids)[i])[j] = counts[i] ? (*centroidsTemp[i])[j] / counts[i] : (*centroidsTemp[i])[j];
      }
    }

    progress_str.sprintf("%d #%d", iter, changed);
    if (iter>0) {
      if (changed_max<changed) changed_max = changed;
      doProgressProc( changed_max-changed, changed_max, progress_str.c_str() );
    } else
      doProgressProc( 0, 100, progress_str.c_str() );

    if (changed == 0) break;
    if ( (max_iterations>0) && (iter>max_iterations) ) break;
    ++iter;
  } // while (fabs(error - old_error) > t)

  // Delete all threads
  for(i=0;i<threads.size();++i) {
    KMeansThread<T> *thread = threads[i];
    delete thread;
  }
  threads.clear(); 

  is_running = false;
}

#endif // #ifndef KM_SINGLE_THREAD


#endif // DIM_KMEANS
