/*****************************************************************************
 YAKMEL (Yet Another K-Means Clustering Library) - HTD Specification
 HTD - Homogenious Texture Descriptor
 This library provides clustering of HTD feature vectors

 the 60 dim HTD feature vector is organized as following:
 vScaleOrientation, i.e. m12 - mean of scale 1 and orientation 2
 30 means: [m11,m12,...m16,m21,m22,...m26,m31,..m36,.......,m51,...m56]
 30 stds:  [s11,s12,...s16,s21,s22,...s26,s31,..s36,.......,s51,...s56]
 
 Defined classes:
   CDimTextureKMeans      - inherited class accepting Texture data

 Author: Dima V. Fedorov <mailto:dima@dimin.net> <http://www.dimin.net/>
 Copyright (c) 2006 Vision Research Lab, UCSB <http://vision.ece.ucsb.edu>

 History:
   04/19/2006 16:20 - First creation

 Ver : 1
*****************************************************************************/

#ifndef DIM_KMEANS_HTD
#define DIM_KMEANS_HTD

//----------------------------------------------------------
// Configuration
//----------------------------------------------------------

// define this to use specific distance tailored for HTD
#define KM_USE_TEXTURE_DISTANCE

// Default distance function:
//   distance_htd_euclidean
//   distance_htd_emd
#define KM_HTD_DISTANCE_FUNCTION distance_htd_euclidean

//----------------------------------------------------------
// Code
//----------------------------------------------------------

#include "kmeans.h"


//****************************************************************************
// CDimTextureKMeans
//****************************************************************************

template <class T>
class CDimTextureKMeans: public CDimThreadedKMeansBase<T> {
public:
	typedef CDimThreadedKMeansBase<T> _MyBase;

  CDimTextureKMeans( int nthreads, int nclasses, int ndimensions );

  void run( const CDimImageList* _images, CDimTileList* _tiles );

  //-----------------------------------------------------------------------
  // analysis part
  //-----------------------------------------------------------------------

  // find the ID of the assigned class closest to the GT class
  // gt_class - id of the ground truth class of interest
  // min_gt_prob - minimum probability of GT class, if 0 then the class_groundtruth value will be used instead
  // min_gt_prob: 0 - FULL, 90 - CLEAR, 100 - PURE
  int getClassClosestToGT( CDimTileList* _tiles, int gt_class, int min_gt_prob );
  
  // precision amd recall
  // gt_class - id of the ground truth class of interest
  // as_class - id of the assigned class of interest (should be retrieved using getClassClosestToGT)
  // min_gt_prob - minimum probability of GT class, if 0 then the class_groundtruth value will be used instead
  // min_gt_prob: 0 - FULL, 90 - CLEAR, 100 - PURE
  void getPrecisionRecall( CDimTileList* _tiles, int gt_class, int as_class, int min_gt_prob, double &precision, double &recall );

protected:
  CDimTileList  *tiles;
  CDimImageList *images;
  
#ifdef KM_USE_TEXTURE_DISTANCE
  virtual inline double get_distance( CDimFeatureVector<T> *item1, CDimFeatureVector<T> *item2 );  
#endif

private:
  // hide the inherited run method to avoid compiler warnings
  void run( CDimVectorOfFeatures<T> *_data ) { _data; };
};

template<class T>
CDimTextureKMeans<T>::CDimTextureKMeans( int nthreads, int nclasses, int ndimensions )
: CDimThreadedKMeansBase<T>(nthreads, nclasses, ndimensions+0) {
//: CDimThreadedKMeansBase<T>(nthreads, nclasses, ndimensions) {
  images = NULL;
  tiles = NULL;
}

template<class T>
void CDimTextureKMeans<T>::run( const CDimImageList* _images, CDimTileList* _tiles ) {
  if ( (_images == NULL) || (_images->empty()) ) return;
  if ( (_tiles == NULL) || (_tiles->empty()) ) return;

  // assign vectors
  images = (CDimImageList*) _images;
  tiles  = _tiles;

  // convert data
  // we'd rather have fast index access to data points, so copy into internal vector
  data->clear();
  data->reserve( tiles->size() );

  //iterate through data points and assign
  unsigned int i, j;
  for (i=0; i<tiles->size(); ++i) {
    data->add( (*tiles)[i]->texture, TEXTURE_DIMS );
  }

/*

  for (i=0; i<tiles->size(); ++i) {
    double d;
    d = (double) (*tiles)[i]->position_in_img[0] / (double) (*images)[0]->image_size[0];
    //(*(*data)[i])[TEXTURE_DIMS+0] = (texture_type) d;

    d = (double) (*tiles)[i]->position_in_img[1] / (double) (*images)[0]->image_size[1];
    //(*(*data)[i])[TEXTURE_DIMS+1] = (texture_type) d;
    
    /*
    d = (double) (*tiles)[i]->position_in_img[2] / (double) (*images)[0]->image_size[2];
    (*(*data)[i])[TEXTURE_DIMS+2] = (texture_type) d;

    d = (double) (*tiles)[i]->image_id / (double) images->size();
    (*(*data)[i])[TEXTURE_DIMS+3] = (texture_type) d;
    */
 // }

  /*
  for (i=0; i<tiles->size(); ++i) {
    double d;
    d = (double) (*tiles)[i]->position_in_img[0];
    (*(*data)[i])[TEXTURE_DIMS+0] = (texture_type) d;

    d = (double) (*tiles)[i]->position_in_img[1];
    (*(*data)[i])[TEXTURE_DIMS+1] = (texture_type) d;
    /*

    d = (double) (*tiles)[i]->position_in_img[2];
    (*(*data)[i])[TEXTURE_DIMS+2] = (texture_type) d;

    d = (double) (*tiles)[i]->image_id;
    (*(*data)[i])[TEXTURE_DIMS+3] = (texture_type) d;
    */
  //}

  // normalize data points
  //data->normalize_by_std();
  //data->normalize_by_std( 60 );
  //data->normalize_by_std( 61 );
  //data->normalize_by_std( 62 );
  //data->normalize_by_std( 63 );

  // start clustering
  do_clustering();

  // assign classes to data
  for (i=0; i<tiles->size(); ++i) {
    (*tiles)[i]->class_assigned = labels[i];
  }

  // get probabilistic distances to centroids
  // dima 
  compute_perpoint_distance_distr( );
  //sort_classes_by_max_prob( );

  unsigned int min_centr = dim::min( k, (unsigned int) 12 );
  for (i=0; i<tiles->size(); ++i) {
    for (j=0; j<12; ++j) (*tiles)[i]->layers_assigned[j] = 0;

    for (j=0; j<min_centr; ++j) {
      (*tiles)[i]->layers_assigned[j] = (*(*distributions)[i])[j];
    }
  } // for all data points

  data->clear();
}


// gt_class - id of the ground truth class of interest
// min_gt_prob - minimum probability of GT class, if 0 then the class_groundtruth value will be used instead
// min_gt_prob: 0 - FULL, 90 - CLEAR, 100 - PURE

template<class T>
int CDimTextureKMeans<T>::getClassClosestToGT( CDimTileList* _tiles, int gt_class, int min_gt_prob ) {
  unsigned int i;
  int closest_class = -1;
  if ( (_tiles == NULL) || (_tiles->empty()) ) return closest_class;
  tiles  = _tiles;

  std::vector<double> distances(k, 0.0);
  CDimTileDescriptor gt_texture = tiles->getMeanTextureGT( gt_class, min_gt_prob );
  CDimTileDescriptor as_texture;
  CDimFeatureVector<T> item1( TEXTURE_DIMS );
  CDimFeatureVector<T> item2( TEXTURE_DIMS );
  item1.assign(gt_texture.texture, TEXTURE_DIMS );

  for (i=0; i<k; ++i) {
    as_texture = tiles->getMeanTextureAssigned( i );
    item2.assign(as_texture.texture, TEXTURE_DIMS );    
    distances[i] = get_distance( &item1, &item2 );
  }
 
  double min_dist = distances[0];
  closest_class = 0;
  for (i=1; i<k; ++i) {
    double dist = distances[i];
    if (min_dist > dist) {
      min_dist = dist;
      closest_class = i;
    }
  }
  return closest_class;
}


// recall - ratio of the number of relevant records retrieved to the total
// number of relevant records in the database. 
// precision - ratio of the number of relevant records retrieved to the
// total number of irrelevant and relevant records retrieved

// gt_class - id of the ground truth class of interest
// as_class - id of the assigned class of interest (should be retrieved using getClassClosestToGT)
// min_gt_prob - minimum probability of GT class, if 0 then the class_groundtruth value will be used instead
// min_gt_prob: 0 - FULL, 90 - CLEAR, 100 - PURE

template<class T>
void CDimTextureKMeans<T>::getPrecisionRecall( CDimTileList* _tiles, int gt_class, int as_class, int min_gt_prob, double &precision, double &recall ) {
  precision = 0.0;
  recall  = 0.0;
  if ( (_tiles == NULL) || (_tiles->empty()) ) return;
  tiles  = _tiles;

  unsigned int num_retrieved    = 0;
  unsigned int num_retrieved_ok = 0;
  unsigned int num_gt           = 0;

  // get retrieved points
  register bool is_in_gt = false;
  for (unsigned int i=0; i<tiles->size(); ++i ) {

    is_in_gt = false;
    if ( (tiles->at(i)->class_groundtruth == gt_class) &&
         (tiles->at(i)->layers_gt[gt_class] >= min_gt_prob) ) {
      ++num_gt;
      is_in_gt = true;
    }

    if (tiles->at(i)->class_assigned == as_class) {
      ++num_retrieved;
      if (is_in_gt) ++num_retrieved_ok;
    }    
  }

  recall    = (double) num_retrieved_ok / (double) num_gt;
  precision = (double) num_retrieved_ok / (double) num_retrieved;
}

//****************************************************************************
// Distances tailored for HTD
//****************************************************************************

#ifdef KM_USE_TEXTURE_DISTANCE

//----------------------------------------------------------------------------
// Euclidean - default
// Orientation-invariant eucledian distance
// the 60 dim HTD feature vector is organized as following:
// 30 means: [m11,m12,...m16,m21,m22,...m26,m31,..m36,.......,m51,...m56]
// 30 stds:  [s11,s12,...s16,s21,s22,...s26,s31,..s36,.......,s51,...s56]
//----------------------------------------------------------------------------

template<class T> 
inline double distance_htd_euclidean(CDimFeatureVector<T> *item1, CDimFeatureVector<T> *item2, 
                                 CDimFeatureVector<double> *weights, uint num_dims) {
  // first arrange feature vectors according to the dim meaning by minimizing the 
  // difference by rotating one vector in relation to another
  CDimFeatureVector<T> nv(num_dims);
  nv.assign(*item2);
  double min_dist = distance_euclidean(item1, &nv, weights, num_dims);
  
  // iterate over all 6 orientations and choose the one minimizing the difference
  for (unsigned int r=0; r<5; ++r) {
    // try new orientation
    for (unsigned int b=0; b<60; b+=6) nv.shiftLeft(b, 6);

    // compute distance
    double new_dist = distance_euclidean(item1, &nv, weights, num_dims);
    if (new_dist < min_dist) min_dist = new_dist;
  }

  // compute eucledian distance with rotated vector
	return min_dist;
}

//----------------------------------------------------------------------------
// EMD - Earth Movers Distance
// separate distances per orientations and scales for means and stds
//----------------------------------------------------------------------------

template<class T> 
inline double distance_htd_emd(CDimFeatureVector<T> *item1, CDimFeatureVector<T> *item2, 
                           CDimFeatureVector<double> *weights, uint num_dims) {
  CDimFeatureVector<T> i1(6);
  CDimFeatureVector<T> i2(6);
  signature_t s1 = { 6, (void*) &i1, (void*) weights };
  signature_t s2 = { 6, (void*) &i2, (void*) weights };

  double comb_dist = 0.0;
  for (unsigned int i=0; i<10; ++i) {
    i1.assign( item1->begin( )+i*6, item1->begin( )+i*6+6 );
    i2.assign( item2->begin( )+i*6, item2->begin( )+i*6+6 );
	  comb_dist += emd(&s1, &s2, emd_dist<T>, emd_weight, 0, 0);
  }

  return comb_dist;
}


// default is euclidean distance
template<class T>
inline double CDimTextureKMeans<T>::get_distance( CDimFeatureVector<T> *item1, CDimFeatureVector<T> *item2 ) {
  return KM_HTD_DISTANCE_FUNCTION(item1, item2, &weights, num_dims);
}

#endif // KM_USE_TEXTURE_DISTANCE

#endif // DIM_KMEANS_HTD
