/*****************************************************************************
  descriptors for textures, tiles and images

 IMPLEMENTATION

 Author: Dima V. Fedorov <mailto:dima@dimin.net> <http://www.dimin.net/>
 Copyright (c) 2006 Vision Research Lab, UCSB <http://vision.ece.ucsb.edu>

 History:
   04/19/2006 16:20 - First creation

 Ver : 1
*****************************************************************************/

#include "textureDescriptors.h"

#include <cstdlib>
#include <cmath>

#include <string>
#include <vector>
#include <fstream>

#include <xstring.h>

inline int iround(double x) {
  return (int) floor(x + 0.5);
}

//----------------------------------------------------------------------------
// TDimTileDescriptor
//----------------------------------------------------------------------------

CDimTileDescriptor::CDimTileDescriptor()
{
  image_id = -1;
  tile_id = 0;
  tile_size_pixels = 0;
  memset( position_in_img, 0, sizeof(unsigned int)*5 );
  memset( texture, 0, sizeof(texture_type)*TEXTURE_DIMS );
  memset( layers_assigned, 0, sizeof(char)*12 );
  memset( layers_gt, 0, sizeof(char)*12 );
  class_assigned = -1;
  class_groundtruth = -1;
}

CDimTileDescriptor::~CDimTileDescriptor()
{

}

// buffers - the buffer contains:
  // int32  *1   image_id
  // uint32 *1   tile_id
  // uint32 *1   tile_size_pixels
  // uin32  *3   position_in_img - only first 3 positions XYZ
  // int    *1   class_assigned
  // int    *1   class_groundtruth
  // char   *12  layers_assigned - percentage of each layer in the given tile
  // char   *12  layers_gt - ground truth of percentage of each layer in the given tile 
  // float  *60  texture

unsigned int CDimTileDescriptor::bufferSize()
{
  return ( sizeof(int32)*8 + sizeof(texture_type)*TEXTURE_DIMS + sizeof(char)*24 );
}

unsigned char* CDimTileDescriptor::toBuffer()
{
  unsigned char *buf = new unsigned char [ bufferSize() ];
  if (buf == NULL) return NULL;
  
  unsigned int offset = 0;
  uint32 *ibuf = (uint32 *) buf;
  ibuf[0] = (uint32) image_id;
  ibuf[1] = (uint32) tile_id;
  ibuf[2] = (uint32) tile_size_pixels;
  memcpy( (void*)(ibuf+3), position_in_img, 3*sizeof(unsigned int) );
  ibuf[6] = (uint32) class_assigned;
  ibuf[7] = (uint32) class_groundtruth;
  
  offset = sizeof(int32)*8;
  memcpy( (void*)(buf+offset), layers_assigned, sizeof(char)*12 );
  
  offset += sizeof(char)*12;
  memcpy( (void*)(buf+offset), layers_gt, sizeof(char)*12 );

  offset += sizeof(char)*12;
  memcpy( (void*)(buf+offset), texture, TEXTURE_DIMS*sizeof(texture_type) );

  if (dim::bigendian) {
    dim::swapArrayOfLong(ibuf, 8);
    dim::swapArrayOfFloat((float*)(buf+offset), TEXTURE_DIMS);
  }

  return buf;
}
 
int CDimTileDescriptor::fromBuffer(const unsigned char* buf)
{
  if (buf == NULL) return -1;
  
  unsigned int offset = 0;
  uint32 *ibuf = (uint32 *) buf;
  
  image_id          = (int32) ibuf[0];
  tile_id           = (uint32) ibuf[1];
  tile_size_pixels  = (uint32) ibuf[2];
  memcpy( position_in_img, (void*)(ibuf+3), 3*sizeof(unsigned int) );
  class_assigned    = (uint32) ibuf[6];
  class_groundtruth = (uint32) ibuf[7];
  
  offset = sizeof(int32)*8;
  memcpy( layers_assigned, (void*)(buf+offset), sizeof(char)*12 );
  
  offset += sizeof(char)*12;
  memcpy( layers_gt, (void*)(buf+offset), sizeof(char)*12 );

  offset += sizeof(char)*12;
  memcpy( texture, (void*)(buf+offset), TEXTURE_DIMS*sizeof(texture_type) );

  if (dim::bigendian) {
    dim::swapArrayOfLong(ibuf, 8);
    dim::swapArrayOfFloat((float*)(buf+(sizeof(int32)*8)), TEXTURE_DIMS);
  }
  
  return bufferSize();
}

int CDimTileDescriptor::fromFile(std::ifstream *fs)
{
  if (fs == NULL) return -1;
  const unsigned int bsize = bufferSize();
  unsigned char *buf = new unsigned char [ bsize ];
  if (buf == NULL) return -1;
 
  fs->read((char*)buf, bsize);
  fromBuffer(buf);
  delete buf;
  if (fs->fail()) return -1;
  return (int) bsize;
}

int CDimTileDescriptor::toFile(std::ofstream *fs) {
  if (fs == NULL) return -1;
  const unsigned int bsize = bufferSize();
  unsigned char *buf = toBuffer();
  if (buf == NULL) return -1;
 
  fs->write((char*)buf, bsize);
  delete buf;
  if (fs->fail()) return -1; 
  return (int) bsize;
}

void CDimTileDescriptor::assign(const CDimTileDescriptor &d) {
  (*this) = d;
}

//----------------------------------------------------------------------------
// CDimTileList
//----------------------------------------------------------------------------

int CDimTileList::writeHeader( std::ofstream *fs ) { 
  if (fs == NULL) return -1;
  char header[5] = "TF30";
  fs->write(header, 4);
  if (fs->fail()) return -1; 
  return 1; 
}

int CDimTileList::readHeader( std::ifstream *fs ) { 
  if (fs == NULL) return -1;
  char header[5] = "0000";
  fs->read (header, 4 );
  if (fs->fail()) return -1;
  if (memcmp(header, "TF30", 4) != 0) return -1;
  return 1; 
}

int CDimTileList::getTileIndex(int img_id, unsigned int x, unsigned int y, unsigned int z) { 
  unsigned int i;
  int pos=-1;
  for (i=0; i<size(); ++i) 
    if ( (at(i)->image_id == img_id) && 
         (at(i)->position_in_img[0] == x) && 
         (at(i)->position_in_img[1] == y) && 
         (at(i)->position_in_img[2] == z) ) {
      pos = i;
      break;
    }
  return pos; 
}

// this is a bit cheating, we'll get the maximum value of class_assigned
// presumed: all the classes are assigned consequtively starting with 0
int CDimTileList::getK() { 
  unsigned int i;
  int k=0;

  for (i=0; i<size(); ++i) 
  if (k < at(i)->class_assigned) k = at(i)->class_assigned;

  return k+1; 
}


CDimTileDescriptor CDimTileList::getMeanTextureGT( int gt_class, int min_gt_prob ) { 
  CDimTileDescriptor avgd;
  unsigned int i, j, n=0;
  std::vector <double> sums( TEXTURE_DIMS, 0.0 );
  std::vector <double> layers_as_sums( 12, 0.0 );
  std::vector <double> layers_gt_sums( 12, 0.0 );
  double sums_as_class=0.0;

  avgd.class_groundtruth = gt_class;
  for (i=0; i<size(); ++i) {
    if ( (at(i)->class_groundtruth == gt_class) && 
         (at(i)->layers_gt[gt_class] >= min_gt_prob) ) {
      ++n;
      sums_as_class += at(i)->class_assigned;
      for (j=0; j<TEXTURE_DIMS; ++j) sums[j] += at(i)->texture[j];
      for (j=0; j<12; ++j) layers_gt_sums[j] += at(i)->layers_gt[j];
      for (j=0; j<12; ++j) layers_as_sums[j] += at(i)->layers_assigned[j];
    }
  }

  avgd.class_assigned = iround(sums_as_class / n);
  for (j=0; j<TEXTURE_DIMS; ++j) avgd.texture[j] = (texture_type) (sums[j] / n);  
  for (j=0; j<12; ++j) avgd.layers_gt[j] = iround(layers_gt_sums[j] / n);  
  for (j=0; j<12; ++j) avgd.layers_assigned[j] = iround(layers_as_sums[j] / n); 

  return avgd; 
}

CDimTileDescriptor CDimTileList::getMeanTextureAssigned( int as_class ) { 
  CDimTileDescriptor avgd;
  unsigned int i, j, n=0;
  std::vector <double> sums( TEXTURE_DIMS, 0.0 );
  std::vector <double> layers_as_sums( 12, 0.0 );
  std::vector <double> layers_gt_sums( 12, 0.0 );
  double sums_gt_class=0.0;

  avgd.class_assigned = as_class;
  for (i=0; i<size(); ++i) {
    if (at(i)->class_assigned == as_class) {
      ++n;
      sums_gt_class += at(i)->class_groundtruth;
      for (j=0; j<TEXTURE_DIMS; ++j) sums[j] += at(i)->texture[j];
      for (j=0; j<12; ++j) layers_gt_sums[j] += at(i)->layers_gt[j];
      for (j=0; j<12; ++j) layers_as_sums[j] += at(i)->layers_assigned[j];
    }
  }

  avgd.class_groundtruth = iround(sums_gt_class / n);
  for (j=0; j<TEXTURE_DIMS; ++j) avgd.texture[j] = (texture_type) (sums[j] / n);  
  for (j=0; j<12; ++j) avgd.layers_gt[j] = iround(layers_gt_sums[j] / n);  
  for (j=0; j<12; ++j) avgd.layers_assigned[j] = iround(layers_as_sums[j] / n); 

  return avgd; 
}

//----------------------------------------------------------------------------
// CDimTileListExt - Extended tile list
//----------------------------------------------------------------------------

int CDimTileListExt::getMaxPosInCol(int img_id) { 
  unsigned int i, v=0;
  int pos = getTileIndex(img_id, 0, 0, 0);
  if (pos==-1) return v;

  for (i=pos; i<size(); ++i) {
    if (at(i)->image_id != img_id) break;

    if ( (at(i)->position_in_img[0] == 0) && (v < at(i)->position_in_img[1]) ) {
      v = at(i)->position_in_img[1];
    }
  }

  return v; 
}

int CDimTileListExt::getMaxPosInRow(int img_id) { 
  unsigned int i, v=0;
  int pos = getTileIndex(img_id, 0, 0, 0);
  if (pos==-1) return v;

  for (i=pos; i<size(); ++i) {
    if (at(i)->image_id != img_id) break;

    if ( (at(i)->position_in_img[1] == 0) && (v < at(i)->position_in_img[0]) ) {
      v = at(i)->position_in_img[0];
    }
  }

  return v; 
}

//----------------------------------------------------------------------------
// TDimImageItem
//----------------------------------------------------------------------------

CDimImageItem::CDimImageItem()
{
  image_id = -1;
  memset( image_size, 0, sizeof(unsigned int)*5 );
  memset( pixel_size, 0, sizeof(double)*3 );
}

CDimImageItem::~CDimImageItem()
{

}

int CDimImageItem::fromFile(std::ifstream *fs)
{
  if (fs == NULL) return -1;
  char line[2048] = "\0";
  char name[2048] = "\0";
  float ps[3];

  fs->getline (line, 2048 );
  if (fs->fail()) return -1;

#ifdef HAVE_SECURE_C
  sscanf_s(line, "%d\t%d\t%d\t%d\t%f\t%f\t%f\t%s", &image_id, 
    &image_size[0], &image_size[1], &image_size[2], &ps[0], &ps[1], &ps[2], name, 2048);
#else
  sscanf(line, "%d\t%d\t%d\t%d\t%f\t%f\t%f\t%s", &image_id, 
    &image_size[0], &image_size[1], &image_size[2], &ps[0], &ps[1], &ps[2], name);
#endif

  image_name = name;
  pixel_size[0] = ps[0]; pixel_size[1] = ps[1]; pixel_size[2] = ps[2];

  return 1;
}

int CDimImageItem::toFile(std::ofstream *fs)
{
  if (fs == NULL) return -1;

  xstring line;
  line.sprintf("%d\t%d\t%d\t%d\t%.6f\t%.6f\t%.6f\t%s\n", image_id, image_size[0], image_size[1], image_size[2], 
    pixel_size[0], pixel_size[1], pixel_size[2], image_name.c_str() );
  fs->write(line.c_str(), (std::streamsize) line.size());

  if (fs->fail()) return -1; 
  return 1;
}

//----------------------------------------------------------------------------
// CDimMapItem
//----------------------------------------------------------------------------

CDimMapItem::CDimMapItem()
{
  image_plane_id = -1;
  image_5d_id = -1;
}

int CDimMapItem::fromFile(std::ifstream *fs)
{
  if (fs == NULL) return -1;
  char line[2048] = "\0";
  fs->getline (line, 2048 );
  if (fs->fail()) return -1;
#ifdef HAVE_SECURE_C
  sscanf_s(line, "%d\t%d", &image_plane_id, &image_5d_id);
#else
  sscanf(line, "%d\t%d", &image_plane_id, &image_5d_id);
#endif
  return 1;
}

int CDimMapItem::toFile(std::ofstream *fs)
{
  if (fs == NULL) return -1;

  xstring line;
  line.sprintf( "%d\t%d\n", image_plane_id, image_5d_id );
  fs->write(line.c_str(), (std::streamsize) line.size());

  if (fs->fail()) return -1; 
  return 1;
}


//----------------------------------------------------------------------------
// lists
//----------------------------------------------------------------------------

int CDimImageList::writeHeader( std::ofstream *fs ) { 
  if (fs == NULL) return -1;

  xstring line;
  line.sprintf("img_id\twidth\theight\tsize_z\tpx_sizex\tpx_sizey\tpx_sizez\tfile_name\n" );
  fs->write(line.c_str(), (std::streamsize) line.size());

  if (fs->fail()) return -1; 
  return 1; 
}

int CDimImageList::readHeader( std::ifstream *fs ) { 
  if (fs == NULL) return -1;
  char line[2048] = "\0";
  fs->getline (line, 2048 );
  if (fs->fail()) return -1;
  return 1; 
}


